from enum import StrEnum
from typing import Literal, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from notionary.shared.properties.type import PropertyType
from notionary.shared.typings import JsonDict

# ============================================================================
# Base Model
# ============================================================================


class DataSourceProperty(BaseModel):
    id: str
    name: str
    description: str | None = None
    type: PropertyType


# ============================================================================
# Enums and Shared Config Models
# ============================================================================


class PropertyColor(StrEnum):
    DEFAULT = "default"
    GRAY = "gray"
    BROWN = "brown"
    ORANGE = "orange"
    YELLOW = "yellow"
    GREEN = "green"
    BLUE = "blue"
    PURPLE = "purple"
    PINK = "pink"
    RED = "red"


class NumberFormat(StrEnum):
    NUMBER = "number"
    NUMBER_WITH_COMMAS = "number_with_commas"
    PERCENT = "percent"
    DOLLAR = "dollar"
    AUSTRALIAN_DOLLAR = "australian_dollar"
    CANADIAN_DOLLAR = "canadian_dollar"
    SINGAPORE_DOLLAR = "singapore_dollar"
    EURO = "euro"
    POUND = "pound"
    YEN = "yen"
    RUBLE = "ruble"
    RUPEE = "rupee"
    WON = "won"
    YUAN = "yuan"
    REAL = "real"
    LIRA = "lira"
    RUPIAH = "rupiah"
    FRANC = "franc"
    HONG_KONG_DOLLAR = "hong_kong_dollar"
    NEW_ZEALAND_DOLLAR = "new_zealand_dollar"
    KRONA = "krona"
    NORWEGIAN_KRONE = "norwegian_krone"
    MEXICAN_PESO = "mexican_peso"
    RAND = "rand"
    NEW_TAIWAN_DOLLAR = "new_taiwan_dollar"
    DANISH_KRONE = "danish_krone"
    ZLOTY = "zloty"
    BAHT = "baht"
    FORINT = "forint"
    KORUNA = "koruna"
    SHEKEL = "shekel"
    CHILEAN_PESO = "chilean_peso"
    PHILIPPINE_PESO = "philippine_peso"
    DIRHAM = "dirham"
    COLOMBIAN_PESO = "colombian_peso"
    RIYAL = "riyal"
    RINGGIT = "ringgit"
    LEU = "leu"
    ARGENTINE_PESO = "argentine_peso"
    URUGUAYAN_PESO = "uruguayan_peso"
    PERUVIAN_SOL = "peruvian_sol"


class RelationType(StrEnum):
    SINGLE_PROPERTY = "single_property"
    DUAL_PROPERTY = "dual_property"


class RollupFunction(StrEnum):
    AVERAGE = "average"
    CHECKED = "checked"
    COUNT_PER_GROUP = "count_per_group"
    COUNT = "count"
    COUNT_VALUES = "count_values"
    DATE_RANGE = "date_range"
    EARLIEST_DATE = "earliest_date"
    EMPTY = "empty"
    LATEST_DATE = "latest_date"
    MAX = "max"
    MEDIAN = "median"
    MIN = "min"
    NOT_EMPTY = "not_empty"
    PERCENT_CHECKED = "percent_checked"
    PERCENT_EMPTY = "percent_empty"
    PERCENT_NOT_EMPTY = "percent_not_empty"
    PERCENT_PER_GROUP = "percent_per_group"
    PERCENT_UNCHECKED = "percent_unchecked"
    RANGE = "range"
    UNCHECKED = "unchecked"
    UNIQUE = "unique"
    SHOW_ORIGINAL = "show_original"
    SHOW_UNIQUE = "show_unique"
    SUM = "sum"


class DataSourcePropertyOption(BaseModel):
    id: str
    name: str
    color: PropertyColor
    description: str | None = None


class DataSourceStatusGroup(BaseModel):
    id: str
    name: str
    color: PropertyColor
    option_ids: list[str]


class DataSourceStatusConfig(BaseModel):
    options: list[DataSourcePropertyOption] = Field(default_factory=list)
    groups: list[DataSourceStatusGroup] = Field(default_factory=list)


class DataSourceStatusProperty(DataSourceProperty):
    type: Literal[PropertyType.STATUS] = PropertyType.STATUS
    status: DataSourceStatusConfig = Field(default_factory=DataSourceStatusConfig)

    @property
    def option_names(self) -> list[str]:
        return [option.name for option in self.status.options]

    @property
    def group_names(self) -> list[str]:
        return [group.name for group in self.status.groups]


class DataSourceSelectConfig(BaseModel):
    options: list[DataSourcePropertyOption] = Field(default_factory=list)


class DataSourceSelectProperty(DataSourceProperty):
    type: Literal[PropertyType.SELECT] = PropertyType.SELECT
    select: DataSourceSelectConfig = Field(default_factory=DataSourceSelectConfig)

    @property
    def option_names(self) -> list[str]:
        return [option.name for option in self.select.options]


class DataSourceMultiSelectConfig(BaseModel):
    options: list[DataSourcePropertyOption] = Field(default_factory=list)


class DataSourceMultiSelectProperty(DataSourceProperty):
    type: Literal[PropertyType.MULTI_SELECT] = PropertyType.MULTI_SELECT
    multi_select: DataSourceMultiSelectConfig = Field(
        default_factory=DataSourceMultiSelectConfig
    )

    @property
    def option_names(self) -> list[str]:
        return [option.name for option in self.multi_select.options]


# https://developers.notion.com/reference/property-object
class DataSourceRelationConfig(BaseModel):
    data_source_id: str | None = None
    type: RelationType = RelationType.SINGLE_PROPERTY
    single_property: JsonDict = Field(default_factory=dict)


class DataSourceRelationProperty(DataSourceProperty):
    type: Literal[PropertyType.RELATION] = PropertyType.RELATION
    relation: DataSourceRelationConfig = Field(default_factory=DataSourceRelationConfig)

    @property
    def related_data_source_id(self) -> str | None:
        return self.relation.data_source_id


class DataSourceDateConfig(BaseModel): ...


class DataSourceDateProperty(DataSourceProperty):
    type: Literal[PropertyType.DATE] = PropertyType.DATE
    date: DataSourceDateConfig = Field(default_factory=DataSourceDateConfig)


class DataSourceCreatedTimeConfig(BaseModel): ...


class DataSourceCreatedTimeProperty(DataSourceProperty):
    type: Literal[PropertyType.CREATED_TIME] = PropertyType.CREATED_TIME
    created_time: DataSourceCreatedTimeConfig = Field(
        default_factory=DataSourceCreatedTimeConfig
    )


class DataSourceCreatedByConfig(BaseModel): ...


class DataSourceCreatedByProperty(DataSourceProperty):
    type: Literal[PropertyType.CREATED_BY] = PropertyType.CREATED_BY
    created_by: DataSourceCreatedByConfig = Field(
        default_factory=DataSourceCreatedByConfig
    )


class DataSourceLastEditedTimeConfig(BaseModel): ...


class DataSourceLastEditedTimeProperty(DataSourceProperty):
    type: Literal[PropertyType.LAST_EDITED_TIME] = PropertyType.LAST_EDITED_TIME
    last_edited_time: DataSourceLastEditedTimeConfig = Field(
        default_factory=DataSourceLastEditedTimeConfig
    )


class DataSourceLastEditedByConfig(BaseModel): ...


class DataSourceLastEditedByProperty(DataSourceProperty):
    type: Literal[PropertyType.LAST_EDITED_BY] = PropertyType.LAST_EDITED_BY
    last_edited_by: DataSourceLastEditedByConfig = Field(
        default_factory=DataSourceLastEditedByConfig
    )


class DataSourceLastVisitedTimeConfig(BaseModel): ...


class DataSourceLastVisitedTimeProperty(DataSourceProperty):
    type: Literal[PropertyType.LAST_VISITED_TIME] = PropertyType.LAST_VISITED_TIME
    last_visited_time: DataSourceLastVisitedTimeConfig = Field(
        default_factory=DataSourceLastVisitedTimeConfig
    )


class DataSourceTitleConfig(BaseModel): ...


class DataSourceTitleProperty(DataSourceProperty):
    type: Literal[PropertyType.TITLE] = PropertyType.TITLE
    title: DataSourceTitleConfig = Field(default_factory=DataSourceTitleConfig)


class DataSourceRichTextConfig(BaseModel): ...


class DataSourceRichTextProperty(DataSourceProperty):
    type: Literal[PropertyType.RICH_TEXT] = PropertyType.RICH_TEXT
    rich_text: DataSourceRichTextConfig = Field(
        default_factory=DataSourceRichTextConfig
    )


class DataSourceURLConfig(BaseModel): ...


class DataSourceURLProperty(DataSourceProperty):
    type: Literal[PropertyType.URL] = PropertyType.URL
    url: DataSourceURLConfig = Field(default_factory=DataSourceURLConfig)


class DataSourcePeopleConfig(BaseModel): ...


class DataSourcePeopleProperty(DataSourceProperty):
    type: Literal[PropertyType.PEOPLE] = PropertyType.PEOPLE
    people: DataSourcePeopleConfig = Field(default_factory=DataSourcePeopleConfig)


class DataSourceNumberConfig(BaseModel):
    format: NumberFormat


class DataSourceNumberProperty(DataSourceProperty):
    type: Literal[PropertyType.NUMBER] = PropertyType.NUMBER
    number: DataSourceNumberConfig

    @property
    def number_format(self) -> NumberFormat:
        return self.number.format


class DataSourceCheckboxConfig(BaseModel): ...


class DataSourceCheckboxProperty(DataSourceProperty):
    type: Literal[PropertyType.CHECKBOX] = PropertyType.CHECKBOX
    checkbox: DataSourceCheckboxConfig = Field(default_factory=DataSourceCheckboxConfig)


class DataSourceEmailConfig(BaseModel): ...


class DataSourceEmailProperty(DataSourceProperty):
    type: Literal[PropertyType.EMAIL] = PropertyType.EMAIL
    email: DataSourceEmailConfig = Field(default_factory=DataSourceEmailConfig)


class DataSourcePhoneNumberConfig(BaseModel): ...


class DataSourcePhoneNumberProperty(DataSourceProperty):
    type: Literal[PropertyType.PHONE_NUMBER] = PropertyType.PHONE_NUMBER
    phone_number: DataSourcePhoneNumberConfig = Field(
        default_factory=DataSourcePhoneNumberConfig
    )


class DataSourceFilesConfig(BaseModel): ...


class DataSourceFilesProperty(DataSourceProperty):
    type: Literal[PropertyType.FILES] = PropertyType.FILES
    files: DataSourceFilesConfig = Field(default_factory=DataSourceFilesConfig)


class DataSourceFormulaConfig(BaseModel):
    expression: str


class DataSourceFormulaProperty(DataSourceProperty):
    type: Literal[PropertyType.FORMULA] = PropertyType.FORMULA
    formula: DataSourceFormulaConfig

    @property
    def expression(self) -> str:
        return self.formula.expression


class DataSourceRollupConfig(BaseModel):
    function: RollupFunction
    relation_property_id: str
    relation_property_name: str
    rollup_property_id: str
    rollup_property_name: str


class DataSourceRollupProperty(DataSourceProperty):
    type: Literal[PropertyType.ROLLUP] = PropertyType.ROLLUP
    rollup: DataSourceRollupConfig

    @property
    def rollup_function(self) -> RollupFunction:
        return self.rollup.function


class DataSourceUniqueIdConfig(BaseModel):
    prefix: str | None = None


class DataSourceUniqueIdProperty(DataSourceProperty):
    type: Literal[PropertyType.UNIQUE_ID] = PropertyType.UNIQUE_ID
    unique_id: DataSourceUniqueIdConfig = Field(
        default_factory=DataSourceUniqueIdConfig
    )

    @property
    def prefix(self) -> str | None:
        return self.unique_id.prefix


class DataSourceButtonConfig(BaseModel): ...


class DataSourceButtonProperty(DataSourceProperty):
    type: Literal[PropertyType.BUTTON] = PropertyType.BUTTON
    button: DataSourceButtonConfig = Field(default_factory=DataSourceButtonConfig)


class DataSourceLocationConfig(BaseModel): ...


class DataSourceLocationProperty(DataSourceProperty):
    type: Literal[PropertyType.LOCATION] = PropertyType.LOCATION
    location: DataSourceLocationConfig = Field(default_factory=DataSourceLocationConfig)


class DataSourcePlaceConfig(BaseModel): ...


class DataSourcePlaceProperty(DataSourceProperty):
    type: Literal[PropertyType.PLACE] = PropertyType.PLACE
    place: DataSourcePlaceConfig = Field(default_factory=DataSourcePlaceConfig)


class DataSourceVerificationConfig(BaseModel): ...


class DataSourceVerificationProperty(DataSourceProperty):
    type: Literal[PropertyType.VERIFICATION] = PropertyType.VERIFICATION
    verification: DataSourceVerificationConfig = Field(
        default_factory=DataSourceVerificationConfig
    )


class DataSourceUnknownProperty(BaseModel):
    model_config = ConfigDict(extra="allow")


type AnyDataSourceProperty = (
    DataSourceStatusProperty
    | DataSourceSelectProperty
    | DataSourceMultiSelectProperty
    | DataSourceRelationProperty
    | DataSourceDateProperty
    | DataSourceCreatedTimeProperty
    | DataSourceCreatedByProperty
    | DataSourceLastEditedTimeProperty
    | DataSourceLastEditedByProperty
    | DataSourceLastVisitedTimeProperty
    | DataSourceTitleProperty
    | DataSourceRichTextProperty
    | DataSourceURLProperty
    | DataSourcePeopleProperty
    | DataSourceNumberProperty
    | DataSourceCheckboxProperty
    | DataSourceEmailProperty
    | DataSourcePhoneNumberProperty
    | DataSourceFilesProperty
    | DataSourceFormulaProperty
    | DataSourceRollupProperty
    | DataSourceUniqueIdProperty
    | DataSourceButtonProperty
    | DataSourceLocationProperty
    | DataSourcePlaceProperty
    | DataSourceVerificationProperty
    | DataSourceUnknownProperty
)

DataSourcePropertyT = TypeVar("DataSourcePropertyT", bound=DataSourceProperty)



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.markdown.nodes.container import ContainerNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class HeadingMarkdownNode(ContainerNode):
    MIN_LEVEL = 1
    MAX_LEVEL = 3

    def __init__(
        self,
        text: str,
        level: int = 1,
        children: list[MarkdownNode] | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self.text = text
        self.level = self._validate_level(level)
        self.children = children or []

    @override
    def to_markdown(self) -> str:
        heading_prefix = self._get_heading_prefix()
        result = f"{heading_prefix} {self.text}"
        result += self.render_children()
        return result

    def _validate_level(self, level: int) -> int:
        return max(self.MIN_LEVEL, min(self.MAX_LEVEL, level))

    def _get_heading_prefix(self) -> str:
        delimiter = self._syntax_registry.get_heading_syntax().start_delimiter
        return delimiter * self.level



from notionary.blocks.schemas import BlockCreatePayload
from notionary.page.content.parser.parsers import (
    BlockParsingContext,
    ParentBlockContext,
)
from notionary.page.content.parser.parsers.base import LineParser
from notionary.page.content.parser.post_processing.service import BlockPostProcessor
from notionary.page.content.parser.pre_processsing.service import MarkdownPreProcessor
from notionary.utils.mixins.logging import LoggingMixin


class MarkdownToNotionConverter(LoggingMixin):
    def __init__(
        self,
        line_parser: LineParser,
        pre_processor: MarkdownPreProcessor,
        post_processor: BlockPostProcessor,
    ) -> None:
        self._line_parser = line_parser
        self._pre_processor = pre_processor
        self._post_processor = post_processor

    async def convert(self, markdown_text: str) -> list[BlockCreatePayload]:
        if not markdown_text:
            return []

        markdown_text = self._pre_processor.process(markdown_text)
        all_blocks = await self._process_lines(markdown_text)
        all_blocks = self._post_processor.process(all_blocks)

        return all_blocks

    async def _process_lines(self, text: str) -> list[BlockCreatePayload]:
        lines = text.split("\n")
        result_blocks: list[BlockCreatePayload] = []
        parent_stack: list[ParentBlockContext] = []

        current_line_index = 0
        previous_line_was_empty = False

        while current_line_index < len(lines):
            line = lines[current_line_index]

            context = self._create_line_processing_context(
                line=line,
                lines=lines,
                line_index=current_line_index,
                result_blocks=result_blocks,
                parent_stack=parent_stack,
                is_previous_line_empty=previous_line_was_empty,
            )

            await self._line_parser.handle(context)
            previous_line_was_empty = self._is_processed_line_empty(line)

            current_line_index += 1 + context.lines_consumed

        return result_blocks

    def _create_line_processing_context(
        self,
        line: str,
        lines: list[str],
        line_index: int,
        result_blocks: list[BlockCreatePayload],
        parent_stack: list[ParentBlockContext],
        is_previous_line_empty: bool = False,
    ) -> BlockParsingContext:
        return BlockParsingContext(
            line=line,
            result_blocks=result_blocks,
            parent_stack=parent_stack,
            parse_children_callback=self._process_lines,
            all_lines=lines,
            current_line_index=line_index,
            lines_consumed=0,
            is_previous_line_empty=is_previous_line_empty,
        )

    def _is_processed_line_empty(self, line: str) -> bool:
        return line.strip() == ""



from __future__ import annotations

from collections.abc import Callable
from typing import Self

from notionary.blocks.enums import CodingLanguage
from notionary.page.content.markdown.nodes import (
    AudioMarkdownNode,
    BookmarkMarkdownNode,
    BreadcrumbMarkdownNode,
    BulletedListMarkdownNode,
    CalloutMarkdownNode,
    CodeMarkdownNode,
    ColumnListMarkdownNode,
    ColumnMarkdownNode,
    DividerMarkdownNode,
    EmbedMarkdownNode,
    EquationMarkdownNode,
    FileMarkdownNode,
    HeadingMarkdownNode,
    ImageMarkdownNode,
    MarkdownNode,
    NumberedListMarkdownNode,
    ParagraphMarkdownNode,
    PdfMarkdownNode,
    QuoteMarkdownNode,
    SpaceMarkdownNode,
    TableMarkdownNode,
    TableOfContentsMarkdownNode,
    TodoMarkdownNode,
    ToggleMarkdownNode,
    VideoMarkdownNode,
)
from notionary.page.content.markdown.nodes.container import flatten_children


class MarkdownBuilder:
    def __init__(self) -> None:
        self.children: list[MarkdownNode] = []

    def h1(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        return self._add_heading(text, 1, builder_func)

    def h2(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        return self._add_heading(text, 2, builder_func)

    def h3(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        return self._add_heading(text, 3, builder_func)

    def paragraph(self, text: str) -> Self:
        self.children.append(ParagraphMarkdownNode(text=text))
        return self

    def space(self) -> Self:
        self.children.append(SpaceMarkdownNode())
        return self

    def quote(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        children = self._build_children(builder_func)
        self.children.append(QuoteMarkdownNode(text=text, children=children))
        return self

    def divider(self) -> Self:
        self.children.append(DividerMarkdownNode())
        return self

    def numbered_list(self, items: list[str]) -> Self:
        self.children.append(NumberedListMarkdownNode(texts=items))
        return self

    def numbered_list_item(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        children = self._build_children(builder_func)
        wrapped = flatten_children(children)
        child_nodes = [wrapped] if wrapped else None
        self.children.append(
            NumberedListMarkdownNode(texts=[text], children=child_nodes)
        )
        return self

    def bulleted_list(self, items: list[str]) -> Self:
        self.children.append(BulletedListMarkdownNode(texts=items))
        return self

    def bulleted_list_item(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        children = self._build_children(builder_func)
        wrapped = flatten_children(children)
        child_nodes = [wrapped] if wrapped else None
        self.children.append(
            BulletedListMarkdownNode(texts=[text], children=child_nodes)
        )
        return self

    def todo(
        self,
        text: str,
        checked: bool = False,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        children = self._build_children(builder_func)
        self.children.append(
            TodoMarkdownNode(text=text, checked=checked, children=children)
        )
        return self

    def checked_todo(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        return self.todo(text, checked=True, builder_func=builder_func)

    def unchecked_todo(
        self,
        text: str,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        return self.todo(text, checked=False, builder_func=builder_func)

    def todo_list(self, items: list[str], completed: list[bool] | None = None) -> Self:
        completed = completed or [False] * len(items)
        for i, item in enumerate(items):
            is_done = completed[i] if i < len(completed) else False
            self.children.append(TodoMarkdownNode(text=item, checked=is_done))
        return self

    def callout(self, text: str, emoji: str | None = None) -> Self:
        self.children.append(CalloutMarkdownNode(text=text, emoji=emoji))
        return self

    def callout_with_children(
        self,
        text: str,
        emoji: str | None = None,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None = None,
    ) -> Self:
        children = self._build_children(builder_func)
        self.children.append(
            CalloutMarkdownNode(text=text, emoji=emoji, children=children)
        )
        return self

    def toggle(
        self, title: str, builder_func: Callable[[MarkdownBuilder], MarkdownBuilder]
    ) -> Self:
        children = self._build_children(builder_func)
        self.children.append(ToggleMarkdownNode(title=title, children=children))
        return self

    def image(self, url: str, caption: str | None = None) -> Self:
        self.children.append(ImageMarkdownNode(url=url, caption=caption))
        return self

    def video(self, url: str, caption: str | None = None) -> Self:
        self.children.append(VideoMarkdownNode(url=url, caption=caption))
        return self

    def audio(self, url: str, caption: str | None = None) -> Self:
        self.children.append(AudioMarkdownNode(url=url, caption=caption))
        return self

    def file(self, url: str, caption: str | None = None) -> Self:
        self.children.append(FileMarkdownNode(url=url, caption=caption))
        return self

    def pdf(self, url: str, caption: str | None = None) -> Self:
        self.children.append(PdfMarkdownNode(url=url, caption=caption))
        return self

    def bookmark(
        self, url: str, title: str | None = None, caption: str | None = None
    ) -> Self:
        self.children.append(
            BookmarkMarkdownNode(url=url, title=title, caption=caption)
        )
        return self

    def embed(self, url: str, caption: str | None = None) -> Self:
        self.children.append(EmbedMarkdownNode(url=url, caption=caption))
        return self

    def code(
        self,
        code: str,
        language: CodingLanguage | None = None,
        caption: str | None = None,
    ) -> Self:
        self.children.append(
            CodeMarkdownNode(code=code, language=language, caption=caption)
        )
        return self

    def mermaid(self, diagram: str, caption: str | None = None) -> Self:
        self.children.append(
            CodeMarkdownNode(
                code=diagram, language=CodingLanguage.MERMAID.value, caption=caption
            )
        )
        return self

    def table(self, headers: list[str], rows: list[list[str]]) -> Self:
        self.children.append(TableMarkdownNode(headers=headers, rows=rows))
        return self

    def breadcrumb(self) -> Self:
        self.children.append(BreadcrumbMarkdownNode())
        return self

    def equation(self, expression: str) -> Self:
        self.children.append(EquationMarkdownNode(expression=expression))
        return self

    def table_of_contents(self) -> Self:
        self.children.append(TableOfContentsMarkdownNode())
        return self

    def columns(
        self,
        *builder_funcs: Callable[[MarkdownBuilder], MarkdownBuilder],
        width_ratios: list[float] | None = None,
    ) -> Self:
        columns = self._build_columns(builder_funcs, width_ratios)
        self.children.append(ColumnListMarkdownNode(columns=columns))
        return self

    def build(self) -> str:
        return "\n\n".join(
            child.to_markdown() for child in self.children if child is not None
        )

    def _add_heading(
        self,
        text: str,
        level: int,
        builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None,
    ) -> Self:
        children = self._build_children(builder_func)
        self.children.append(
            HeadingMarkdownNode(text=text, level=level, children=children)
        )
        return self

    def _build_children(
        self, builder_func: Callable[[MarkdownBuilder], MarkdownBuilder] | None
    ) -> list[MarkdownNode]:
        if builder_func is None:
            return []

        builder = MarkdownBuilder()
        builder_func(builder)
        return builder.children

    def _build_columns(
        self,
        builder_funcs: tuple[Callable[[MarkdownBuilder], MarkdownBuilder], ...],
        width_ratios: list[float] | None,
    ) -> list[ColumnMarkdownNode]:
        columns = []
        for i, builder_func in enumerate(builder_funcs):
            width_ratio = width_ratios[i] if width_ratios else None
            children = self._build_children(builder_func)
            columns.append(
                ColumnMarkdownNode(children=children, width_ratio=width_ratio)
            )
        return columns



import math
from typing import override

from notionary.page.content.parser.pre_processsing.handlers.port import PreProcessor
from notionary.page.content.syntax.definition import (
    MarkdownGrammar,
    SyntaxDefinitionRegistry,
)
from notionary.utils.decorators import time_execution_sync
from notionary.utils.mixins.logging import LoggingMixin


class IndentationNormalizer(PreProcessor, LoggingMixin):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
        markdown_grammar: MarkdownGrammar | None = None,
    ) -> None:
        super().__init__()
        self._syntax_registry = syntax_registry or SyntaxDefinitionRegistry()
        self._markdown_grammar = markdown_grammar or MarkdownGrammar()

        self._spaces_per_nesting_level = self._markdown_grammar.spaces_per_nesting_level
        self._code_block_start_delimiter = (
            self._syntax_registry.get_code_syntax().start_delimiter
        )

    @override
    @time_execution_sync()
    def process(self, markdown_text: str) -> str:
        if self._is_empty(markdown_text):
            return ""

        normalized = self._normalize_to_markdown_indentation(markdown_text)

        if normalized != markdown_text:
            self.logger.warning(
                "Corrected non-standard indentation. Check the result for formatting errors and use consistent indentation in the source."
            )

        return normalized

    def _is_empty(self, text: str) -> bool:
        return not text

    def _normalize_to_markdown_indentation(self, markdown_text: str) -> str:
        lines = markdown_text.split("\n")
        processed_lines = []
        inside_code_block = False

        for line in lines:
            if self._is_code_fence(line):
                inside_code_block = not inside_code_block
                processed_lines.append(line)
            elif inside_code_block:
                processed_lines.append(line)
            else:
                processed_lines.append(self._normalize_to_standard_indentation(line))

        return "\n".join(processed_lines)

    def _is_code_fence(self, line: str) -> bool:
        return line.lstrip().startswith(self._code_block_start_delimiter)

    def _normalize_to_standard_indentation(self, line: str) -> str:
        if self._is_blank_line(line):
            return ""

        indentation_level = self._round_to_nearest_indentation_level(line)
        content = self._extract_content(line)

        return self._build_indented_line(indentation_level, content)

    def _is_blank_line(self, line: str) -> bool:
        return not line.strip()

    def _round_to_nearest_indentation_level(self, line: str) -> int:
        leading_spaces = self._count_leading_spaces(line)
        return math.ceil(leading_spaces / self._spaces_per_nesting_level)

    def _count_leading_spaces(self, line: str) -> int:
        return len(line) - len(line.lstrip())

    def _extract_content(self, line: str) -> str:
        return line.lstrip()

    def _build_indented_line(self, level: int, content: str) -> str:
        standard_indent = self._create_standard_indent(level)
        return standard_indent + content

    def _create_standard_indent(self, level: int) -> str:
        spaces = level * self._spaces_per_nesting_level
        return " " * spaces



from .api import (
    NotionApiError,
    NotionAuthenticationError,
    NotionConnectionError,
    NotionRateLimitError,
    NotionResourceNotFoundError,
    NotionServerError,
    NotionValidationError,
)
from .base import NotionaryException
from .block_parsing import (
    InsufficientColumnsError,
    InvalidColumnRatioSumError,
    UnsupportedVideoFormatError,
)
from .data_source import DataSourcePropertyNotFound, DataSourcePropertyTypeError
from .file_upload import (
    FileSizeException,
    NoFileExtensionException,
    UnsupportedFileTypeException,
)
from .properties import (
    AccessPagePropertyWithoutDataSourceError,
    PagePropertyNotFoundError,
    PagePropertyTypeError,
)
from .search import DatabaseNotFound, DataSourceNotFound, EntityNotFound, PageNotFound

__all__ = [
    "AccessPagePropertyWithoutDataSourceError",
    "DataSourceNotFound",
    "DataSourcePropertyNotFound",
    "DataSourcePropertyTypeError",
    "DatabaseNotFound",
    "EntityNotFound",
    "FileSizeException",
    "InsufficientColumnsError",
    "InvalidColumnRatioSumError",
    "NoFileExtensionException",
    "NotionApiError",
    "NotionAuthenticationError",
    "NotionConnectionError",
    "NotionRateLimitError",
    "NotionResourceNotFoundError",
    "NotionServerError",
    "NotionValidationError",
    "NotionaryException",
    "PageNotFound",
    "PagePropertyNotFoundError",
    "PagePropertyTypeError",
    "UnsupportedFileTypeException",
    "UnsupportedVideoFormatError",
]



from pydantic import BaseModel, field_validator, model_serializer

from notionary.file_upload.schemas import FileUploadStatus


class FileUploadQuery(BaseModel):
    status: FileUploadStatus | None = None
    archived: bool | None = None

    page_size_limit: int | None = None
    total_results_limit: int | None = None

    @field_validator("page_size_limit")
    @classmethod
    def validate_page_size(cls, value: int | None) -> int | None:
        if value is None:
            return None
        return max(1, min(value, 100))

    @field_validator("total_results_limit")
    @classmethod
    def validate_total_results(cls, value: int | None) -> int:
        if value is None:
            return 100
        return max(1, value)

    @model_serializer
    def serialize_model(self) -> dict[str, str | bool | None]:
        result = {}

        if self.status is not None:
            result["status"] = self.status

        if self.archived is not None:
            result["archived"] = self.archived

        return result



from __future__ import annotations

from abc import ABC, abstractmethod

from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class BlockRenderer(ABC):
    def __init__(self, syntax_registry: SyntaxDefinitionRegistry | None = None) -> None:
        self._syntax_registry = syntax_registry or SyntaxDefinitionRegistry()
        self._next_handler: BlockRenderer | None = None

    def set_next(self, handler: BlockRenderer) -> BlockRenderer:
        self._next_handler = handler
        return handler

    async def handle(self, context: MarkdownRenderingContext) -> None:
        if self._can_handle(context.block):
            await self._process(context)
        elif self._next_handler:
            await self._next_handler.handle(context)

    @abstractmethod
    def _can_handle(self, block: Block) -> bool:
        pass

    @abstractmethod
    async def _process(self, context: MarkdownRenderingContext) -> None:
        pass



from notionary.data_source.schemas import DataSourceDto
from notionary.http.client import NotionHttpClient


class DataSourceClient(NotionHttpClient):
    def __init__(self, timeout: int = 30) -> None:
        super().__init__(timeout)

    async def get_data_source(self, data_source_id: str) -> DataSourceDto:
        response = await self.get(f"data_sources/{data_source_id}")
        return DataSourceDto.model_validate(response)



from abc import ABC, abstractmethod

from notionary.shared.entity.schemas import EntityResponseDto, NotionEntityUpdateDto
from notionary.shared.models.file import ExternalFile, FileUploadFile
from notionary.shared.models.icon import EmojiIcon, Icon


class EntityMetadataUpdateClient(ABC):
    @abstractmethod
    async def patch_metadata(
        self, updated_data: NotionEntityUpdateDto
    ) -> EntityResponseDto: ...

    async def patch_emoji_icon(self, emoji: str) -> EntityResponseDto:
        icon = EmojiIcon(emoji=emoji)
        update_dto = NotionEntityUpdateDto(icon=icon)
        return await self.patch_metadata(update_dto)

    async def patch_external_icon(self, icon_url: str) -> EntityResponseDto:
        icon = ExternalFile.from_url(icon_url)
        return await self._patch_icon(icon)

    async def patch_icon_from_file_upload(
        self, file_upload_id: str
    ) -> EntityResponseDto:
        icon = FileUploadFile.from_id(id=file_upload_id)
        return await self._patch_icon(icon)

    async def _patch_icon(self, icon: Icon) -> EntityResponseDto:
        update_dto = NotionEntityUpdateDto(icon=icon)
        return await self.patch_metadata(update_dto)

    async def remove_icon(self) -> None:
        update_dto = NotionEntityUpdateDto(icon=None)
        return await self.patch_metadata(update_dto)

    async def patch_external_cover(self, cover_url: str) -> EntityResponseDto:
        cover = ExternalFile.from_url(cover_url)
        return await self._patch_cover(cover)

    async def patch_cover_from_file_upload(
        self, file_upload_id: str
    ) -> EntityResponseDto:
        cover = FileUploadFile.from_id(id=file_upload_id)
        return await self._patch_cover(cover)

    async def _patch_cover(self, cover: Icon) -> EntityResponseDto:
        update_dto = NotionEntityUpdateDto(cover=cover)
        return await self.patch_metadata(update_dto)

    async def remove_cover(self) -> None:
        update_dto = NotionEntityUpdateDto(cover=None)
        return await self.patch_metadata(update_dto)

    async def move_to_trash(self) -> EntityResponseDto:
        update_dto = NotionEntityUpdateDto(in_trash=True)
        return await self.patch_metadata(update_dto)

    async def restore_from_trash(self) -> EntityResponseDto:
        update_dto = NotionEntityUpdateDto(in_trash=False)
        return await self.patch_metadata(update_dto)



from typing import override

from notionary.blocks.enums import BlockType
from notionary.blocks.schemas import (
    BlockCreatePayload,
    CreateColumnListBlock,
    CreateColumnListData,
)
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class ColumnListParser(LineParser):
    def __init__(self, syntax_registry: SyntaxDefinitionRegistry) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_column_list_syntax()

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        return self._is_column_list_start(context)

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        if self._is_column_list_start(context):
            await self._process_column_list(context)

    def _is_column_list_start(self, context: BlockParsingContext) -> bool:
        return self._syntax.regex_pattern.match(context.line) is not None

    async def _process_column_list(self, context: BlockParsingContext) -> None:
        block = self._create_column_list_block()
        await self._populate_columns(block, context)
        context.result_blocks.append(block)

    def _create_column_list_block(self) -> CreateColumnListBlock:
        column_list_data = CreateColumnListData(children=[])
        return CreateColumnListBlock(column_list=column_list_data)

    async def _populate_columns(
        self, block: CreateColumnListBlock, context: BlockParsingContext
    ) -> None:
        parent_indent_level = context.get_line_indentation_level()
        child_lines = self._collect_children_allowing_empty_lines(
            context, parent_indent_level
        )

        if not child_lines:
            return

        column_blocks = await self._parse_column_children(child_lines, context)
        block.column_list.children = column_blocks
        context.lines_consumed = len(child_lines)

    async def _parse_column_children(
        self, child_lines: list[str], context: BlockParsingContext
    ) -> list:
        stripped_lines = context.strip_indentation_level(child_lines, levels=1)
        child_markdown = "\n".join(stripped_lines)
        parsed_blocks = await context.parse_nested_markdown(child_markdown)
        return self._extract_column_blocks(parsed_blocks)

    def _collect_children_allowing_empty_lines(
        self, context: BlockParsingContext, parent_indent_level: int
    ) -> list[str]:
        child_lines = []
        expected_child_indent = parent_indent_level + 1
        remaining_lines = context.get_remaining_lines()

        for line in remaining_lines:
            if self._should_include_as_child(line, expected_child_indent, context):
                child_lines.append(line)
            else:
                break

        return child_lines

    def _should_include_as_child(
        self, line: str, expected_indent: int, context: BlockParsingContext
    ) -> bool:
        if not line.strip():
            return True

        line_indent = context.get_line_indentation_level(line)
        return line_indent >= expected_indent

    def _extract_column_blocks(self, blocks: list[BlockCreatePayload]) -> list:
        return [block for block in blocks if self._is_valid_column_block(block)]

    def _is_valid_column_block(self, block: BlockCreatePayload) -> bool:
        return block.type == BlockType.COLUMN and block.column is not None



"""
Fixed API limits and recommendations from Notion.
These values should not be changed, as they are specified by the API.

Source: https://developers.notion.com/reference/file-uploads
"""

_MB = 1024 * 1024

NOTION_SINGLE_PART_MAX_SIZE: int = 20 * _MB
NOTION_MAX_FILENAME_BYTES: int = 900

NOTION_MULTI_PART_CHUNK_SIZE_MIN: int = 5 * _MB
NOTION_MULTI_PART_CHUNK_SIZE_MAX: int = 20 * _MB

NOTION_RECOMMENDED_CHUNK_SIZE: int = 10 * _MB



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Never

from notionary.blocks.rich_text.rich_text_markdown_converter import (
    convert_rich_text_to_markdown,
)
from notionary.exceptions.properties import (
    AccessPagePropertyWithoutDataSourceError,
    PagePropertyNotFoundError,
    PagePropertyTypeError,
)
from notionary.page.properties.client import PagePropertyHttpClient
from notionary.page.properties.schemas import (
    PageCheckboxProperty,
    PageCreatedTimeProperty,
    PageDateProperty,
    PageEmailProperty,
    PageMultiSelectProperty,
    PageNumberProperty,
    PagePeopleProperty,
    PagePhoneNumberProperty,
    PageProperty,
    PagePropertyT,
    PageRelationProperty,
    PageRichTextProperty,
    PageSelectProperty,
    PageStatusProperty,
    PageTitleProperty,
    PageURLProperty,
)
from notionary.shared.models.parent import ParentType

if TYPE_CHECKING:
    from notionary import NotionDataSource


class PagePropertyHandler:
    def __init__(
        self,
        properties: dict[str, PageProperty],
        parent_type: ParentType,
        page_url: str,
        page_property_http_client: PagePropertyHttpClient,
        parent_data_source: str,
    ) -> None:
        self._properties = properties
        self._parent_type = parent_type
        self._page_url = page_url
        self._property_http_client = page_property_http_client
        self._parent_data_source_id = parent_data_source
        self._parent_data_source: NotionDataSource | None = None
        self._data_source_loaded = False

    # =========================================================================
    # Reader Methods
    # =========================================================================

    def get_value_of_status_property(self, name: str) -> str | None:
        status_property = self._get_typed_property_or_raise(name, PageStatusProperty)
        return status_property.status.name if status_property.status else None

    def get_value_of_select_property(self, name: str) -> str | None:
        select_property = self._get_typed_property_or_raise(name, PageSelectProperty)
        return select_property.select.name if select_property.select else None

    async def get_value_of_title_property(self, name: str) -> str:
        title_property = self._get_typed_property_or_raise(name, PageTitleProperty)
        return await convert_rich_text_to_markdown(title_property.title)

    def get_values_of_people_property(self, property_name: str) -> list[str]:
        people_prop = self._get_typed_property_or_raise(
            property_name, PagePeopleProperty
        )
        return [person.name for person in people_prop.people if person.name]

    def get_value_of_created_time_property(self, name: str) -> str | None:
        created_time_property = self._get_typed_property_or_raise(
            name, PageCreatedTimeProperty
        )
        return created_time_property.created_time

    async def get_values_of_relation_property(self, name: str) -> list[str]:
        from notionary import NotionPage

        relation_property = self._get_typed_property_or_raise(
            name, PageRelationProperty
        )
        relation_page_ids = [rel.id for rel in relation_property.relation]
        notion_pages = [
            await NotionPage.from_id(page_id) for page_id in relation_page_ids
        ]
        return [page.title for page in notion_pages if page]

    def get_values_of_multiselect_property(self, name: str) -> list[str]:
        multiselect_property = self._get_typed_property_or_raise(
            name, PageMultiSelectProperty
        )
        return [option.name for option in multiselect_property.multi_select]

    def get_value_of_url_property(self, name: str) -> str | None:
        url_property = self._get_typed_property_or_raise(name, PageURLProperty)
        return url_property.url

    def get_value_of_number_property(self, name: str) -> float | None:
        number_property = self._get_typed_property_or_raise(name, PageNumberProperty)
        return number_property.number

    def get_value_of_checkbox_property(self, name: str) -> bool:
        checkbox_property = self._get_typed_property_or_raise(
            name, PageCheckboxProperty
        )
        return checkbox_property.checkbox

    def get_value_of_date_property(self, name: str) -> str | None:
        date_property = self._get_typed_property_or_raise(name, PageDateProperty)
        return date_property.date.start if date_property.date else None

    async def get_value_of_rich_text_property(self, name: str) -> str:
        rich_text_property = self._get_typed_property_or_raise(
            name, PageRichTextProperty
        )
        return await convert_rich_text_to_markdown(rich_text_property.rich_text)

    def get_value_of_email_property(self, name: str) -> str | None:
        email_property = self._get_typed_property_or_raise(name, PageEmailProperty)
        return email_property.email

    def get_value_of_phone_number_property(self, name: str) -> str | None:
        phone_property = self._get_typed_property_or_raise(
            name, PagePhoneNumberProperty
        )
        return phone_property.phone_number

    # =========================================================================
    # Options Getters
    # =========================================================================

    async def get_select_options_by_property_name(
        self, property_name: str
    ) -> list[str]:
        data_source = await self._get_parent_data_source_or_raise()
        return data_source.get_select_options_by_property_name(property_name)

    async def get_multi_select_options_by_property_name(
        self, property_name: str
    ) -> list[str]:
        data_source = await self._get_parent_data_source_or_raise()
        return data_source.get_multi_select_options_by_property_name(property_name)

    async def get_status_options_by_property_name(
        self, property_name: str
    ) -> list[str]:
        data_source = await self._get_parent_data_source_or_raise()
        return data_source.get_status_options_by_property_name(property_name)

    async def get_relation_options_by_property_name(
        self, property_name: str
    ) -> list[str]:
        data_source = await self._get_parent_data_source_or_raise()
        return await data_source.get_relation_options_by_property_name(property_name)

    async def get_options_for_property_by_name(self, property_name: str) -> list[str]:
        data_source = await self._get_parent_data_source_or_raise()
        return await data_source.get_options_for_property_by_name(property_name)

    async def get_schema_description(self, property_name: str) -> str:
        data_source = await self._get_parent_data_source_or_raise()
        return await data_source.get_schema_description(property_name)

    # =========================================================================
    # Writer Methods
    # =========================================================================

    async def set_title_property(self, title: str) -> None:
        title_property_name = self._extract_title_property_name()

        self._get_typed_property_or_raise(title_property_name, PageTitleProperty)
        updated_page = await self._property_http_client.patch_title(
            title_property_name, title
        )
        self._properties = updated_page.properties

    def _extract_title_property_name(self) -> str | None:
        if not self._properties:
            return None

        return next(
            (
                key
                for key, prop in self._properties.items()
                if isinstance(prop, PageTitleProperty)
            ),
            None,
        )

    async def set_rich_text_property(self, property_name: str, text: str) -> None:
        self._get_typed_property_or_raise(property_name, PageRichTextProperty)
        updated_page = await self._property_http_client.patch_rich_text_property(
            property_name, text
        )
        self._properties = updated_page.properties

    async def set_url_property(self, property_name: str, url: str) -> None:
        self._get_typed_property_or_raise(property_name, PageURLProperty)
        updated_page = await self._property_http_client.patch_url_property(
            property_name, url
        )
        self._properties = updated_page.properties

    async def set_email_property(self, property_name: str, email: str) -> None:
        self._get_typed_property_or_raise(property_name, PageEmailProperty)
        updated_page = await self._property_http_client.patch_email_property(
            property_name, email
        )
        self._properties = updated_page.properties

    async def set_phone_number_property(
        self, property_name: str, phone_number: str
    ) -> None:
        self._get_typed_property_or_raise(property_name, PagePhoneNumberProperty)
        updated_page = await self._property_http_client.patch_phone_property(
            property_name, phone_number
        )
        self._properties = updated_page.properties

    async def set_number_property(
        self, property_name: str, number: int | float
    ) -> None:
        self._get_typed_property_or_raise(property_name, PageNumberProperty)
        updated_page = await self._property_http_client.patch_number_property(
            property_name, number
        )
        self._properties = updated_page.properties

    async def set_checkbox_property(self, property_name: str, checked: bool) -> None:
        self._get_typed_property_or_raise(property_name, PageCheckboxProperty)
        updated_page = await self._property_http_client.patch_checkbox_property(
            property_name, checked
        )
        self._properties = updated_page.properties

    async def set_date_property(
        self, property_name: str, date_value: str | dict
    ) -> None:
        self._get_typed_property_or_raise(property_name, PageDateProperty)
        updated_page = await self._property_http_client.patch_date_property(
            property_name, date_value
        )
        self._properties = updated_page.properties

    async def set_select_property_by_option_name(
        self, property_name: str, option_name: str
    ) -> None:
        self._get_typed_property_or_raise(property_name, PageSelectProperty)
        updated_page = await self._property_http_client.patch_select_property(
            property_name, option_name
        )
        self._properties = updated_page.properties

    async def set_multi_select_property_by_option_names(
        self, property_name: str, option_names: list[str]
    ) -> None:
        self._get_typed_property_or_raise(property_name, PageMultiSelectProperty)
        updated_page = await self._property_http_client.patch_multi_select_property(
            property_name, option_names
        )
        self._properties = updated_page.properties

    async def set_status_property_by_option_name(
        self, property_name: str, status: str
    ) -> None:
        self._get_typed_property_or_raise(property_name, PageStatusProperty)
        updated_page = await self._property_http_client.patch_status_property(
            property_name, status
        )
        self._properties = updated_page.properties

    async def set_relation_property_by_page_titles(
        self, property_name: str, page_titles: list[str]
    ) -> None:
        self._get_typed_property_or_raise(property_name, PageRelationProperty)
        relation_ids = await self._convert_page_titles_to_ids(page_titles)
        updated_page = await self._property_http_client.patch_relation_property(
            property_name, relation_ids
        )
        self._properties = updated_page.properties

    async def _ensure_data_source_loaded(self) -> None:
        from notionary import NotionDataSource

        if self._data_source_loaded:
            return

        self._parent_data_source = (
            await NotionDataSource.from_id(self._parent_data_source_id)
            if self._parent_data_source_id
            else None
        )
        self._data_source_loaded = True

    async def _get_parent_data_source_or_raise(self) -> NotionDataSource:
        await self._ensure_data_source_loaded()

        if not self._parent_data_source:
            raise AccessPagePropertyWithoutDataSourceError(self._parent_type)
        return self._parent_data_source

    def _get_typed_property_or_raise(
        self, name: str, property_type: type[PagePropertyT]
    ) -> PagePropertyT:
        prop = self._properties.get(name)

        if prop is None:
            self._handle_prop_not_found(name)

        if not isinstance(prop, property_type):
            self._handle_incorrect_type(name, type(prop))

        return prop

    def _handle_prop_not_found(self, name: str) -> Never:
        raise PagePropertyNotFoundError(
            property_name=name,
            page_url=self._page_url,
            available_properties=list(self._properties.keys()),
        )

    def _handle_incorrect_type(self, property_name: str, actual_type: type) -> Never:
        raise PagePropertyTypeError(
            property_name=property_name,
            actual_type=actual_type.__name__,
        )

    async def _convert_page_titles_to_ids(self, page_titles: list[str]) -> list[str]:
        from notionary import NotionPage

        if not page_titles:
            return []

        pages = await asyncio.gather(
            *[NotionPage.from_title(title=title) for title in page_titles]
        )

        return [page.id for page in pages if page]



from typing import override

from notionary.blocks.enums import BlockType
from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer


class TableRowHandler(BlockRenderer):
    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.TABLE_ROW

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        """Table rows are internally handled by table as the structure supports it"""
        pass



from typing import override

from notionary.database.schemas import NotionDatabaseDto
from notionary.http.client import NotionHttpClient
from notionary.shared.entity.entity_metadata_update_client import (
    EntityMetadataUpdateClient,
)
from notionary.shared.entity.schemas import NotionEntityUpdateDto


class DatabaseMetadataUpdateClient(NotionHttpClient, EntityMetadataUpdateClient):
    def __init__(self, database_id: str, timeout: int = 30) -> None:
        super().__init__(timeout)
        self._database_id = database_id

    @override
    async def patch_metadata(
        self, updated_data: NotionEntityUpdateDto
    ) -> NotionDatabaseDto:
        updated_data_dict = updated_data.model_dump(
            exclude_unset=True, exclude_none=True
        )

        response_dict = await self.patch(
            f"databases/{self._database_id}", data=updated_data_dict
        )
        return NotionDatabaseDto.model_validate(response_dict)



from enum import StrEnum
from typing import Self

from pydantic import BaseModel

from notionary.blocks.enums import BlockColor


class RichTextType(StrEnum):
    TEXT = "text"
    MENTION = "mention"
    EQUATION = "equation"


class MentionType(StrEnum):
    USER = "user"
    PAGE = "page"
    DATABASE = "database"
    DATASOURCE = "data_source"
    DATE = "date"
    LINK_PREVIEW = "link_preview"
    TEMPLATE_MENTION = "template_mention"


class TemplateMentionType(StrEnum):
    USER = "template_mention_user"
    DATE = "template_mention_date"


class TextAnnotations(BaseModel):
    bold: bool = False
    italic: bool = False
    strikethrough: bool = False
    underline: bool = False
    code: bool = False
    color: BlockColor | None = None


class LinkObject(BaseModel):
    url: str


class TextContent(BaseModel):
    content: str
    link: LinkObject | None = None


class EquationObject(BaseModel):
    expression: str


class MentionUserRef(BaseModel):
    id: str  # Notion user id


class MentionPageRef(BaseModel):
    id: str


class MentionDatabaseRef(BaseModel):
    id: str


class MentionDataSourceRef(BaseModel):
    id: str


class MentionLinkPreview(BaseModel):
    url: str


class MentionDate(BaseModel):
    # entspricht Notion date object (start Pflicht, end/time_zone optional)
    start: str  # ISO 8601 date or datetime
    end: str | None = None
    time_zone: str | None = None


class MentionTemplateMention(BaseModel):
    # Notion hat zwei Template-Mention-Typen
    type: TemplateMentionType


class MentionObject(BaseModel):
    type: MentionType
    user: MentionUserRef | None = None
    page: MentionPageRef | None = None
    database: MentionDatabaseRef | None = None
    data_source: MentionDataSourceRef | None = None
    date: MentionDate | None = None
    link_preview: MentionLinkPreview | None = None
    template_mention: MentionTemplateMention | None = None


class RichText(BaseModel):
    type: RichTextType = RichTextType.TEXT

    text: TextContent | None = None
    annotations: TextAnnotations | None = None
    plain_text: str = ""
    href: str | None = None

    mention: MentionObject | None = None

    equation: EquationObject | None = None

    @classmethod
    def from_plain_text(cls, content: str, **ann) -> Self:
        return cls(
            type=RichTextType.TEXT,
            text=TextContent(content=content),
            annotations=TextAnnotations(**ann) if ann else TextAnnotations(),
            plain_text=content,
        )

    @classmethod
    def for_caption(cls, content: str) -> Self:
        return cls(
            type=RichTextType.TEXT,
            text=TextContent(content=content),
            annotations=None,
            plain_text=content,
        )

    @classmethod
    def for_code_block(cls, content: str) -> Self:
        # keine annotations setzen → Notion Code-Highlight bleibt an
        return cls.for_caption(content)

    @classmethod
    def for_link(cls, content: str, url: str, **ann) -> Self:
        return cls(
            type=RichTextType.TEXT,
            text=TextContent(content=content, link=LinkObject(url=url)),
            annotations=TextAnnotations(**ann) if ann else TextAnnotations(),
            plain_text=content,
        )

    @classmethod
    def mention_user(cls, user_id: str) -> Self:
        return cls(
            type=RichTextType.MENTION,
            mention=MentionObject(
                type=MentionType.USER, user=MentionUserRef(id=user_id)
            ),
            annotations=TextAnnotations(),
        )

    @classmethod
    def mention_page(cls, page_id: str) -> Self:
        return cls(
            type=RichTextType.MENTION,
            mention=MentionObject(
                type=MentionType.PAGE, page=MentionPageRef(id=page_id)
            ),
            annotations=TextAnnotations(),
        )

    @classmethod
    def mention_database(cls, database_id: str) -> Self:
        return cls(
            type=RichTextType.MENTION,
            mention=MentionObject(
                type=MentionType.DATABASE, database=MentionDatabaseRef(id=database_id)
            ),
            annotations=TextAnnotations(),
        )

    @classmethod
    def mention_data_source(cls, data_source_id: str) -> Self:
        return cls(
            type=RichTextType.MENTION,
            mention=MentionObject(
                type=MentionType.DATASOURCE,
                data_source=MentionDataSourceRef(id=data_source_id),
            ),
            annotations=TextAnnotations(),
        )

    @classmethod
    def equation_inline(cls, expression: str) -> Self:
        return cls(
            type=RichTextType.EQUATION,
            equation=EquationObject(expression=expression),
            annotations=TextAnnotations(),
            plain_text=expression,
        )



import re
from collections.abc import Callable
from dataclasses import dataclass
from re import Match
from typing import ClassVar

from notionary.blocks.rich_text.models import (
    MentionType,
    RichText,
    RichTextType,
    TextAnnotations,
)
from notionary.blocks.rich_text.name_id_resolver import (
    DatabaseNameIdResolver,
    DataSourceNameIdResolver,
    NameIdResolver,
    PageNameIdResolver,
    PersonNameIdResolver,
)
from notionary.blocks.rich_text.rich_text_patterns import RichTextPatterns
from notionary.blocks.schemas import BlockColor


@dataclass
class PatternMatch:
    match: Match
    handler: Callable[[Match], RichText | list[RichText]]
    position: int

    @property
    def matched_text(self) -> str:
        return self.match.group(0)

    @property
    def end_position(self) -> int:
        return self.position + len(self.matched_text)


@dataclass
class PatternHandler:
    pattern: str
    handler: Callable[[Match], RichText | list[RichText]]


class MarkdownRichTextConverter:
    VALID_COLORS: ClassVar[set[str]] = {color.value for color in BlockColor}

    def __init__(
        self,
        *,
        page_resolver: NameIdResolver | None = None,
        database_resolver: NameIdResolver | None = None,
        data_source_resolver: NameIdResolver | None = None,
        person_resolver: NameIdResolver | None = None,
    ):
        self.page_resolver = page_resolver or PageNameIdResolver()
        self.database_resolver = database_resolver or DatabaseNameIdResolver()
        self.data_source_resolver = data_source_resolver or DataSourceNameIdResolver()
        self.person_resolver = person_resolver or PersonNameIdResolver()
        self.format_handlers = self._setup_format_handlers()

    def _setup_format_handlers(self) -> list[PatternHandler]:
        return [
            PatternHandler(RichTextPatterns.BOLD, self._handle_bold_pattern),
            PatternHandler(RichTextPatterns.ITALIC, self._handle_italic_pattern),
            PatternHandler(
                RichTextPatterns.ITALIC_UNDERSCORE, self._handle_italic_pattern
            ),
            PatternHandler(RichTextPatterns.UNDERLINE, self._handle_underline_pattern),
            PatternHandler(
                RichTextPatterns.STRIKETHROUGH, self._handle_strikethrough_pattern
            ),
            PatternHandler(RichTextPatterns.CODE, self._handle_code_pattern),
            PatternHandler(RichTextPatterns.LINK, self._handle_link_pattern),
            PatternHandler(
                RichTextPatterns.INLINE_EQUATION, self._handle_equation_pattern
            ),
            PatternHandler(RichTextPatterns.COLOR, self._handle_color_pattern),
            PatternHandler(
                RichTextPatterns.PAGE_MENTION, self._handle_page_mention_pattern
            ),
            PatternHandler(
                RichTextPatterns.DATABASE_MENTION, self._handle_database_mention_pattern
            ),
            PatternHandler(
                RichTextPatterns.DATASOURCE_MENTION,
                self._handle_data_source_mention_pattern,
            ),
            PatternHandler(
                RichTextPatterns.USER_MENTION, self._handle_user_mention_pattern
            ),
        ]

    async def to_rich_text(self, text: str) -> list[RichText]:
        if not text:
            return []
        return await self._split_text_into_segments(text)

    async def _split_text_into_segments(self, text: str) -> list[RichText]:
        segments: list[RichText] = []
        remaining_text = text

        while remaining_text:
            pattern_match = self._find_earliest_pattern_match(remaining_text)

            if not pattern_match:
                segments.append(RichText.from_plain_text(remaining_text))
                break

            plain_text_before = remaining_text[: pattern_match.position]
            if plain_text_before:
                segments.append(RichText.from_plain_text(plain_text_before))

            pattern_result = await self._process_pattern_match(pattern_match)
            self._add_pattern_result_to_segments(segments, pattern_result)

            remaining_text = remaining_text[pattern_match.end_position :]

        return segments

    def _find_earliest_pattern_match(self, text: str) -> PatternMatch | None:
        """Find the pattern that appears earliest in the text."""
        earliest_match = None
        earliest_position = len(text)

        for pattern_handler in self.format_handlers:
            match = re.search(pattern_handler.pattern, text)
            if match and match.start() < earliest_position:
                earliest_match = PatternMatch(
                    match=match, handler=pattern_handler.handler, position=match.start()
                )
                earliest_position = match.start()

        return earliest_match

    async def _process_pattern_match(
        self, pattern_match: PatternMatch
    ) -> RichText | list[RichText]:
        handler_method = pattern_match.handler

        if self._is_async_handler(handler_method):
            return await handler_method(pattern_match.match)
        else:
            return handler_method(pattern_match.match)

    def _is_async_handler(self, handler_method: Callable) -> bool:
        async_handlers = {
            self._handle_page_mention_pattern,
            self._handle_database_mention_pattern,
            self._handle_data_source_mention_pattern,
            self._handle_color_pattern,  # Color pattern needs async for recursive parsing
            self._handle_user_mention_pattern,
        }
        return handler_method in async_handlers

    def _add_pattern_result_to_segments(
        self, segments: list[RichText], pattern_result: RichText | list[RichText]
    ) -> None:
        if isinstance(pattern_result, list):
            segments.extend(pattern_result)
        elif pattern_result:
            segments.append(pattern_result)

    async def _handle_color_pattern(self, match: Match) -> list[RichText]:
        color, content = match.group(1).lower(), match.group(2)

        if color not in self.VALID_COLORS:
            return [RichText.from_plain_text(f"({match.group(1)}:{content})")]

        parsed_segments = await self._split_text_into_segments(content)

        colored_segments = []
        for segment in parsed_segments:
            if segment.type == RichTextType.TEXT:
                colored_segment = self._apply_color_to_text_segment(segment, color)
                colored_segments.append(colored_segment)
            else:
                colored_segments.append(segment)

        return colored_segments

    def _apply_color_to_text_segment(self, segment: RichText, color: str) -> RichText:
        if segment.type != RichTextType.TEXT:
            return segment

        has_link = segment.text and segment.text.link

        if has_link:
            return self._apply_color_to_link_segment(segment, color)
        else:
            return self._apply_color_to_plain_text_segment(segment, color)

    def _apply_color_to_link_segment(self, segment: RichText, color: str) -> RichText:
        formatting = self._extract_formatting_attributes(segment.annotations)

        return RichText.for_link(
            segment.plain_text, segment.text.link.url, color=color, **formatting
        )

    def _apply_color_to_plain_text_segment(
        self, segment: RichText, color: str
    ) -> RichText:
        if segment.type != RichTextType.TEXT:
            return segment

        formatting = self._extract_formatting_attributes(segment.annotations)

        return RichText.from_plain_text(segment.plain_text, color=color, **formatting)

    def _extract_formatting_attributes(
        self, annotations: TextAnnotations
    ) -> dict[str, bool]:
        if not annotations:
            return {
                "bold": False,
                "italic": False,
                "strikethrough": False,
                "underline": False,
                "code": False,
            }

        return {
            "bold": annotations.bold,
            "italic": annotations.italic,
            "strikethrough": annotations.strikethrough,
            "underline": annotations.underline,
            "code": annotations.code,
        }

    async def _handle_page_mention_pattern(self, match: Match) -> RichText:
        identifier = match.group(1)
        return await self._create_mention_or_fallback(
            identifier=identifier,
            resolve_func=self.page_resolver.resolve_name_to_id,
            create_mention_func=RichText.mention_page,
            mention_type=MentionType.PAGE,
        )

    async def _handle_database_mention_pattern(self, match: Match) -> RichText:
        identifier = match.group(1)
        return await self._create_mention_or_fallback(
            identifier=identifier,
            resolve_func=self.database_resolver.resolve_name_to_id,
            create_mention_func=RichText.mention_database,
            mention_type=MentionType.DATABASE,
        )

    async def _handle_data_source_mention_pattern(self, match: Match) -> RichText:
        identifier = match.group(1)
        return await self._create_mention_or_fallback(
            identifier=identifier,
            resolve_func=self.data_source_resolver.resolve_name_to_id,
            create_mention_func=RichText.mention_data_source,
            mention_type=MentionType.DATASOURCE,
        )

    async def _handle_user_mention_pattern(self, match: Match) -> RichText:
        identifier = match.group(1)
        return await self._create_mention_or_fallback(
            identifier=identifier,
            resolve_func=self.person_resolver.resolve_name_to_id,
            create_mention_func=RichText.mention_user,
            mention_type=MentionType.USER,
        )

    async def _create_mention_or_fallback(
        self,
        identifier: str,
        resolve_func: Callable[[str], str | None],
        create_mention_func: Callable[[str], RichText],
        mention_type: MentionType,
    ) -> RichText:
        try:
            resolved_id = await resolve_func(identifier)

            if resolved_id:
                return create_mention_func(resolved_id)
            else:
                return self._create_unresolved_mention_fallback(
                    identifier, mention_type
                )

        except Exception:
            # If resolution throws an error, fallback to plain text
            return self._create_unresolved_mention_fallback(identifier, mention_type)

    def _create_unresolved_mention_fallback(
        self, identifier: str, mention_type: MentionType
    ) -> RichText:
        fallback_text = f"@{mention_type.value}[{identifier}]"
        return RichText.for_caption(fallback_text)

    def _handle_bold_pattern(self, match: Match) -> RichText:
        return RichText.from_plain_text(match.group(1), bold=True)

    def _handle_italic_pattern(self, match: Match) -> RichText:
        return RichText.from_plain_text(match.group(1), italic=True)

    def _handle_underline_pattern(self, match: Match) -> RichText:
        return RichText.from_plain_text(match.group(1), underline=True)

    def _handle_strikethrough_pattern(self, match: Match) -> RichText:
        return RichText.from_plain_text(match.group(1), strikethrough=True)

    def _handle_code_pattern(self, match: Match) -> RichText:
        return RichText.from_plain_text(match.group(1), code=True)

    def _handle_link_pattern(self, match: Match) -> RichText:
        link_text, url = match.group(1), match.group(2)
        return RichText.for_link(link_text, url)

    def _handle_equation_pattern(self, match: Match) -> RichText:
        expression = match.group(1)
        return RichText.equation_inline(expression)



from .grammar import MarkdownGrammar
from .models import EnclosedSyntaxDefinition, SimpleSyntaxDefinition, SyntaxDefinition
from .registry import SyntaxDefinitionRegistry

__all__ = [
    "EnclosedSyntaxDefinition",
    "MarkdownGrammar",
    "SimpleSyntaxDefinition",
    "SyntaxDefinition",
    "SyntaxDefinitionRegistry",
]



from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.database.schemas import (
    NotionDatabaseDto,
    NotionDatabaseUpdateDto,
)
from notionary.http.client import NotionHttpClient


class NotionDatabaseHttpClient(NotionHttpClient):
    def __init__(self, database_id: str, timeout: int = 30) -> None:
        super().__init__(timeout)
        self._database_id = database_id

    async def get_database(self) -> NotionDatabaseDto:
        response = await self.get(f"databases/{self._database_id}")
        return NotionDatabaseDto.model_validate(response)

    async def patch_database(
        self, update_database_dto: NotionDatabaseUpdateDto
    ) -> NotionDatabaseDto:
        update_database_dto_dict = update_database_dto.model_dump(exclude_none=True)

        response = await self.patch(
            f"databases/{self._database_id}", data=update_database_dto_dict
        )
        return NotionDatabaseDto.model_validate(response)

    async def update_database_title(self, title: str) -> NotionDatabaseDto:
        from notionary.blocks.rich_text.markdown_rich_text_converter import (
            MarkdownRichTextConverter,
        )

        markdown_rich_text_formatter = MarkdownRichTextConverter()
        database_rich_text = await markdown_rich_text_formatter.to_rich_text(title)

        database_title_update_dto = NotionDatabaseUpdateDto(title=database_rich_text)
        return await self.patch_database(database_title_update_dto)

    async def update_database_description(self, description: str) -> str:
        from notionary.blocks.rich_text.markdown_rich_text_converter import (
            MarkdownRichTextConverter,
        )

        markdown_to_rich_text_converter = MarkdownRichTextConverter()
        rich_text_description = await markdown_to_rich_text_converter.to_rich_text(
            description
        )

        database_description_update_dto = NotionDatabaseUpdateDto(
            description=rich_text_description
        )
        update_database_response = await self.patch_database(
            database_description_update_dto
        )

        rich_text_to_markdown_converter = RichTextToMarkdownConverter()
        return await rich_text_to_markdown_converter.to_markdown(
            update_database_response.description
        )



from datetime import datetime


def parse_date(date_str: str) -> str:
    supported_formats = _get_supported_date_formats()

    for date_format in supported_formats:
        parsed_date = _try_parse_date_with_format(date_str, date_format)
        if _date_was_successfully_parsed(parsed_date):
            return _convert_to_iso_format(parsed_date)

    _raise_invalid_date_format_error(date_str)


def _get_supported_date_formats() -> list[str]:
    return [
        "%Y-%m-%d",  # ISO: 2024-12-31
        "%d.%m.%Y",  # German: 31.12.2024
        "%m/%d/%Y",  # US with slash: 12/31/2024
        "%m-%d-%Y",  # US with dash: 12-31-2024
        "%d/%m/%Y",  # Day first with slash: 31/12/2024
        "%d-%m-%Y",  # Day first with dash: 31-12-2024
        "%d-%b-%Y",  # Short month: 31-Dec-2024
        "%d %b %Y",  # Short month with space: 31 Dec 2024
        "%d-%B-%Y",  # Full month: 31-December-2024
        "%d %B %Y",  # Full month with space: 31 December 2024
    ]


def _try_parse_date_with_format(date_str: str, date_format: str) -> datetime | None:
    try:
        return datetime.strptime(date_str, date_format)
    except ValueError:
        return None


def _date_was_successfully_parsed(parsed_date: datetime | None) -> bool:
    return parsed_date is not None


def _convert_to_iso_format(parsed_date: datetime) -> str:
    return parsed_date.strftime("%Y-%m-%d")


def _raise_invalid_date_format_error(date_str: str) -> None:
    error_message = (
        f"Invalid date format: '{date_str}'. "
        f"Supported formats: YYYY-MM-DD, DD.MM.YYYY, MM/DD/YYYY, DD/MM/YYYY, "
        f"DD-Mon-YYYY, DD Month YYYY"
    )
    raise ValueError(error_message)



from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.schemas import (
    BlockColor,
    BlockCreatePayload,
    BlockType,
    CreateHeading1Block,
    CreateHeading2Block,
    CreateHeading3Block,
    CreateHeadingBlock,
    CreateHeadingData,
)
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class HeadingParser(LineParser):
    MIN_HEADING_LEVEL = 1
    MAX_HEADING_LEVEL = 3

    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry,
        rich_text_converter: MarkdownRichTextConverter,
    ) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_heading_syntax()
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._syntax.regex_pattern.match(context.line) is not None

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = await self._create_heading_block(context.line)
        if not block:
            return

        await self._process_nested_children(block, context)
        context.result_blocks.append(block)

    async def _process_nested_children(
        self, block: CreateHeadingBlock, context: BlockParsingContext
    ) -> None:
        parent_indent_level = context.get_line_indentation_level()
        child_lines = context.collect_indented_child_lines(parent_indent_level)

        if not child_lines:
            return

        child_lines = self._remove_trailing_empty_lines(child_lines)

        if not child_lines:
            return

        self._set_heading_toggleable(block, True)

        stripped_lines = context.strip_indentation_level(child_lines, levels=1)
        child_markdown = "\n".join(stripped_lines)

        child_blocks = await context.parse_nested_markdown(child_markdown)
        self._set_heading_children(block, child_blocks)

        context.lines_consumed = len(child_lines)

    def _set_heading_toggleable(
        self, block: CreateHeadingBlock, is_toggleable: bool
    ) -> None:
        if block.type == BlockType.HEADING_1:
            block.heading_1.is_toggleable = is_toggleable
        elif block.type == BlockType.HEADING_2:
            block.heading_2.is_toggleable = is_toggleable
        elif block.type == BlockType.HEADING_3:
            block.heading_3.is_toggleable = is_toggleable

    def _set_heading_children(
        self, block: CreateHeadingBlock, children: list[BlockCreatePayload]
    ) -> None:
        if block.type == BlockType.HEADING_1:
            block.heading_1.children = children
        elif block.type == BlockType.HEADING_2:
            block.heading_2.children = children
        elif block.type == BlockType.HEADING_3:
            block.heading_3.children = children

    def _remove_trailing_empty_lines(self, lines: list[str]) -> list[str]:
        while lines and not lines[-1].strip():
            lines.pop()
        return lines

    async def _create_heading_block(self, line: str) -> CreateHeadingBlock | None:
        match = self._syntax.regex_pattern.match(line)
        if not match:
            return None

        level = len(match.group(1))
        content = match.group(2).strip()

        if not self._is_valid_heading(level, content):
            return None

        heading_data = await self._build_heading_data(content)
        return self._create_heading_block_by_level(level, heading_data)

    def _is_valid_heading(self, level: int, content: str) -> bool:
        return self.MIN_HEADING_LEVEL <= level <= self.MAX_HEADING_LEVEL and bool(
            content
        )

    async def _build_heading_data(self, content: str) -> CreateHeadingData:
        rich_text = await self._rich_text_converter.to_rich_text(content)
        return CreateHeadingData(
            rich_text=rich_text,
            color=BlockColor.DEFAULT,
            is_toggleable=False,
            children=[],
        )

    def _create_heading_block_by_level(
        self, level: int, heading_data: CreateHeadingData
    ) -> CreateHeadingBlock:
        if level == 1:
            return CreateHeading1Block(heading_1=heading_data)
        elif level == 2:
            return CreateHeading2Block(heading_2=heading_data)
        else:
            return CreateHeading3Block(heading_3=heading_data)



import asyncio

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.comments.client import CommentClient
from notionary.comments.factory import CommentFactory
from notionary.comments.models import Comment


class CommentService:
    def __init__(
        self,
        client: CommentClient | None = None,
        factory: CommentFactory | None = None,
        markdown_rich_text_converter: MarkdownRichTextConverter | None = None,
    ) -> None:
        self.client = client or CommentClient()
        self.factory = factory or CommentFactory()
        self._converter = markdown_rich_text_converter or MarkdownRichTextConverter()

    async def list_all_comments_for_page(self, page_id: str) -> list[Comment]:
        comment_dtos = [dto async for dto in self.client.iter_comments(page_id)]

        comments = await asyncio.gather(
            *(self.factory.create_from_dto(dto) for dto in comment_dtos)
        )
        return comments

    async def create_comment_on_page(self, page_id: str, text: str) -> Comment:
        rich_text = await self._converter.to_rich_text(text)
        comment_dto = await self.client.create_comment_for_page(
            rich_text=rich_text, page_id=page_id
        )
        return await self.factory.create_from_dto(comment_dto)

    async def reply_to_discussion_by_id(self, discussion_id: str, text: str) -> Comment:
        rich_text = await self._converter.to_rich_text(text)
        comment_dto = await self.client.create_comment_for_discussion(
            rich_text=rich_text, discussion_id=discussion_id
        )
        return await self.factory.create_from_dto(comment_dto)



from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.post_processing.service import (
    MarkdownRenderingPostProcessor,
)
from notionary.page.content.renderer.renderers import BlockRenderer
from notionary.utils.mixins.logging import LoggingMixin


class NotionToMarkdownConverter(LoggingMixin):
    def __init__(
        self,
        renderer_chain: BlockRenderer,
        post_processor: MarkdownRenderingPostProcessor,
    ) -> None:
        self._renderer_chain = renderer_chain
        self._post_processor = post_processor

    async def convert(self, blocks: list[Block], indent_level: int = 0) -> str:
        if not blocks:
            return ""

        rendered_block_parts = []
        current_block_index = 0

        while current_block_index < len(blocks):
            context = self._create_rendering_context(
                blocks, current_block_index, indent_level
            )
            await self._renderer_chain.handle(context)

            if context.markdown_result:
                rendered_block_parts.append(context.markdown_result)

            current_block_index += 1

        result = self._join_rendered_blocks(rendered_block_parts, indent_level)
        result = self._post_processor.process(result)

        return result

    def _create_rendering_context(
        self, blocks: list[Block], block_index: int, indent_level: int
    ) -> MarkdownRenderingContext:
        block = blocks[block_index]
        return MarkdownRenderingContext(
            block=block,
            indent_level=indent_level,
            convert_children_callback=self.convert,
        )

    def _join_rendered_blocks(
        self, rendered_parts: list[str], indent_level: int
    ) -> str:
        separator = "\n\n" if indent_level == 0 else "\n"
        return separator.join(rendered_parts)



from __future__ import annotations

from enum import StrEnum
from typing import Self

from pydantic import (
    BaseModel,
    ValidationInfo,
    field_validator,
    model_serializer,
    model_validator,
)

from notionary.shared.properties.type import PropertyType
from notionary.shared.typings import JsonDict


class FieldType(StrEnum):
    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    DATE = "date"
    DATETIME = "datetime"
    ARRAY = "array"
    RELATION = "relation"
    PEOPLE = "people"


class StringOperator(StrEnum):
    EQUALS = "equals"
    DOES_NOT_EQUAL = "does_not_equal"
    CONTAINS = "contains"
    DOES_NOT_CONTAIN = "does_not_contain"
    STARTS_WITH = "starts_with"
    ENDS_WITH = "ends_with"
    IS_EMPTY = "is_empty"
    IS_NOT_EMPTY = "is_not_empty"


class NumberOperator(StrEnum):
    EQUALS = "equals"
    DOES_NOT_EQUAL = "does_not_equal"
    GREATER_THAN = "greater_than"
    GREATER_THAN_OR_EQUAL_TO = "greater_than_or_equal_to"
    LESS_THAN = "less_than"
    LESS_THAN_OR_EQUAL_TO = "less_than_or_equal_to"
    IS_EMPTY = "is_empty"
    IS_NOT_EMPTY = "is_not_empty"


class BooleanOperator(StrEnum):
    IS_TRUE = "is_true"
    IS_FALSE = "is_false"


class SelectOperator(StrEnum):
    EQUALS = "equals"
    DOES_NOT_EQUAL = "does_not_equal"
    IS_EMPTY = "is_empty"
    IS_NOT_EMPTY = "is_not_empty"


class DateOperator(StrEnum):
    EQUALS = "equals"
    BEFORE = "before"
    AFTER = "after"
    ON_OR_BEFORE = "on_or_before"
    ON_OR_AFTER = "on_or_after"
    IS_EMPTY = "is_empty"
    IS_NOT_EMPTY = "is_not_empty"


class ArrayOperator(StrEnum):
    CONTAINS = "contains"
    DOES_NOT_CONTAIN = "does_not_contain"
    IS_EMPTY = "is_empty"
    IS_NOT_EMPTY = "is_not_empty"


class LogicalOperator(StrEnum):
    AND = "and"
    OR = "or"


class SortDirection(StrEnum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class TimestampType(StrEnum):
    CREATED_TIME = "created_time"
    LAST_EDITED_TIME = "last_edited_time"


class TimeUnit(StrEnum):
    DAYS = "days"
    WEEKS = "weeks"
    MONTHS = "months"
    YEARS = "years"


type Operator = (
    StringOperator | NumberOperator | BooleanOperator | DateOperator | ArrayOperator
)
type FilterValue = str | int | float | bool | list[str | int | float]


class FilterCondition(BaseModel):
    field: str
    field_type: FieldType
    operator: Operator
    value: FilterValue | None = None
    time_value: int | None = None
    time_unit: TimeUnit | None = None

    @model_validator(mode="after")
    def validate_operator_and_value(self) -> Self:
        self._validate_no_value_operators()
        self._validate_value_required_operators()
        self._validate_value_type_matches_field_type()
        return self

    def _validate_no_value_operators(self) -> None:
        no_value_ops = {
            StringOperator.IS_EMPTY,
            StringOperator.IS_NOT_EMPTY,
            NumberOperator.IS_EMPTY,
            NumberOperator.IS_NOT_EMPTY,
            BooleanOperator.IS_TRUE,
            BooleanOperator.IS_FALSE,
            DateOperator.IS_EMPTY,
            DateOperator.IS_NOT_EMPTY,
            ArrayOperator.IS_EMPTY,
            ArrayOperator.IS_NOT_EMPTY,
        }
        if self.operator in no_value_ops and self.value is not None:
            raise ValueError(f"Operator '{self.operator}' does not expect a value")

    def _validate_value_required_operators(self) -> None:
        operators_to_skip = {
            StringOperator.IS_EMPTY,
            StringOperator.IS_NOT_EMPTY,
            NumberOperator.IS_EMPTY,
            NumberOperator.IS_NOT_EMPTY,
            BooleanOperator.IS_TRUE,
            BooleanOperator.IS_FALSE,
            DateOperator.IS_EMPTY,
            DateOperator.IS_NOT_EMPTY,
            ArrayOperator.IS_EMPTY,
            ArrayOperator.IS_NOT_EMPTY,
        }

        is_skipped_operator = self.operator in operators_to_skip
        if not is_skipped_operator and self.value is None:
            raise ValueError(f"Operator '{self.operator}' requires a value")

    def _validate_value_type_matches_field_type(self) -> None:
        if self.value is None:
            return

        if self.field_type == FieldType.STRING:
            self._ensure_value_is_string()
        elif self.field_type == FieldType.NUMBER:
            self._ensure_value_is_number()
        elif self.field_type == FieldType.BOOLEAN:
            self._ensure_value_is_boolean()
        elif self.field_type in (
            FieldType.DATE,
            FieldType.DATETIME,
        ) or self.field_type in (
            FieldType.ARRAY,
            FieldType.RELATION,
            FieldType.PEOPLE,
        ):
            self._ensure_value_is_string()

    def _ensure_value_is_string(self) -> None:
        if not isinstance(self.value, str):
            raise ValueError(
                f"Value for field type '{self.field_type}' must be a string, got {type(self.value).__name__}"
            )

    def _ensure_value_is_number(self) -> None:
        if not isinstance(self.value, (int, float)):
            raise ValueError(
                f"Value for field type '{self.field_type}' must be a number (int or float), "
                f"got {type(self.value).__name__}"
            )

    def _ensure_value_is_boolean(self) -> None:
        if not isinstance(self.value, bool):
            raise ValueError(
                f"Value for field type '{self.field_type}' must be a boolean, got {type(self.value).__name__}"
            )

    @field_validator("operator")
    @classmethod
    def validate_operator_for_field_type(
        cls,
        value: Operator,
        info: ValidationInfo,
    ) -> Operator:
        if "field_type" not in info.data:
            return value

        field_type: FieldType = info.data["field_type"]
        operator_value = value if isinstance(value, str) else value.value

        if not cls._is_operator_valid_for_field_type(operator_value, field_type):
            raise ValueError(
                f"Operator '{operator_value}' is not valid for field type '{field_type}'"
            )

        return value

    @staticmethod
    def _is_operator_valid_for_field_type(operator: str, field_type: FieldType) -> bool:
        valid_operators: dict[FieldType, list[str]] = {
            FieldType.STRING: [op.value for op in StringOperator],
            FieldType.NUMBER: [op.value for op in NumberOperator],
            FieldType.BOOLEAN: [op.value for op in BooleanOperator],
            FieldType.DATE: [op.value for op in DateOperator],
            FieldType.DATETIME: [op.value for op in DateOperator],
            FieldType.ARRAY: [op.value for op in ArrayOperator],
            FieldType.RELATION: [op.value for op in ArrayOperator],
            FieldType.PEOPLE: [op.value for op in ArrayOperator],
        }

        return operator in valid_operators.get(field_type, [])


class OrGroupMarker(BaseModel):
    conditions: list[FilterCondition]


type InternalFilterCondition = FilterCondition | OrGroupMarker


class PropertyFilter(BaseModel):
    property: str
    property_type: PropertyType
    operator: Operator
    value: FilterValue | None = None

    @model_validator(mode="after")
    def validate_value_type(self) -> Self:
        if self.value is None:
            return self

        if self.property_type in (
            PropertyType.PEOPLE,
            PropertyType.RELATION,
        ) and not isinstance(self.value, str):
            raise ValueError(
                f"Value for property type '{self.property_type.value}' must be a string, "
                f"got {type(self.value).__name__}"
            )

        return self

    @model_serializer
    def serialize_model(self) -> JsonDict:
        property_type_str = self.property_type.value
        operator_str = self.operator.value
        filter_value = self.value

        if isinstance(self.operator, BooleanOperator):
            operator_str = "equals"
            filter_value = self.operator == BooleanOperator.IS_TRUE

        return {
            "property": self.property,
            property_type_str: {
                operator_str: filter_value if filter_value is not None else True
            },
        }


class CompoundFilter(BaseModel):
    operator: LogicalOperator
    filters: list[PropertyFilter | CompoundFilter]

    @model_serializer
    def serialize_model(self) -> JsonDict:
        operator_str = self.operator.value
        return {operator_str: [f.model_dump() for f in self.filters]}


type NotionFilter = PropertyFilter | CompoundFilter


class PropertySort(BaseModel):
    property: str
    direction: SortDirection


class TimestampSort(BaseModel):
    timestamp: TimestampType
    direction: SortDirection


type NotionSort = PropertySort | TimestampSort


class DataSourceQueryParams(BaseModel):
    filter: NotionFilter | None = None
    sorts: list[NotionSort] | None = None
    page_size: int | None = None

    total_results_limit: int | None = None

    @model_serializer
    def to_api_params(self) -> JsonDict:
        result: JsonDict = {}

        if self.filter is not None:
            result["filter"] = self.filter.model_dump()

        if self.sorts is not None and len(self.sorts) > 0:
            result["sorts"] = [sort.model_dump() for sort in self.sorts]

        if self.page_size is not None:
            result["page_size"] = self.page_size

        return result



from typing import override

from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer


class BreadcrumbRenderer(BlockRenderer):
    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.BREADCRUMB

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        syntax = self._syntax_registry.get_breadcrumb_syntax()
        breadcrumb_markdown = syntax.start_delimiter

        if context.indent_level > 0:
            breadcrumb_markdown = context.indent_text(breadcrumb_markdown)

        context.markdown_result = breadcrumb_markdown



from .builder import DataSourceQueryBuilder
from .resolver import QueryResolver
from .schema import DataSourceQueryParams

__all__ = [
    "DataSourceQueryBuilder",
    "DataSourceQueryParams",
    "QueryResolver",
]



from typing import Self

from notionary.file_upload.query.models import FileUploadQuery
from notionary.file_upload.schemas import FileUploadStatus


class FileUploadQueryBuilder:
    def __init__(self, query: FileUploadQuery | None = None):
        self._query = query or FileUploadQuery()

    def with_status(self, status: FileUploadStatus) -> Self:
        self._query.status = status
        return self

    def with_uploaded_status_only(self) -> Self:
        self._query.status = FileUploadStatus.UPLOADED
        return self

    def with_pending_status_only(self) -> Self:
        self._query.status = FileUploadStatus.PENDING
        return self

    def with_failed_status_only(self) -> Self:
        self._query.status = FileUploadStatus.FAILED
        return self

    def with_expired_status_only(self) -> Self:
        self._query.status = FileUploadStatus.EXPIRED
        return self

    def with_archived(self, archived: bool) -> Self:
        self._query.archived = archived
        return self

    def with_page_size_limit(self, page_size_limit: int) -> Self:
        self._query.page_size_limit = self._validate_page_size_limit(page_size_limit)
        return self

    def _validate_page_size_limit(self, value: int) -> int:
        if not (1 <= value <= 100):
            raise ValueError(f"page_size_limit must be between 1 and 100, got {value}")
        return value

    def with_total_results_limit(self, total_results_limit: int) -> Self:
        self._query.total_results_limit = self._validate_total_results_limit(
            total_results_limit
        )
        return self

    def _validate_total_results_limit(self, value: int) -> int:
        if not (1 <= value <= 100):
            raise ValueError(
                f"total_results_limit must be between 1 and 100, got {value}"
            )
        return value

    def build(self) -> FileUploadQuery:
        return self._query



from collections.abc import AsyncGenerator, Callable, Coroutine
from typing import Any

from pydantic import BaseModel


class PaginatedResponse(BaseModel):
    results: list[Any]
    has_more: bool
    next_cursor: str | None


async def _fetch_data(
    api_call: Callable[..., Coroutine[Any, Any, PaginatedResponse]],
    total_results_limit: int | None = None,
    **kwargs,
) -> AsyncGenerator[PaginatedResponse]:
    next_cursor: str | None = None
    has_more: bool = True
    total_fetched: int = 0
    api_page_size: int = kwargs.get("page_size", 100)

    while has_more and _should_continue_fetching(total_results_limit, total_fetched):
        request_params = _build_request_params(kwargs, next_cursor)
        response = await api_call(**request_params)

        limited_results = _apply_result_limit(
            response.results, total_results_limit, total_fetched
        )
        total_fetched += len(limited_results)

        yield _create_limited_response(response, limited_results, api_page_size)

        if _has_reached_limit(total_results_limit, total_fetched):
            break

        has_more = response.has_more
        next_cursor = response.next_cursor


def _should_continue_fetching(total_limit: int | None, total_fetched: int) -> bool:
    if total_limit is None:
        return True
    return total_fetched < total_limit


def _build_request_params(
    base_kwargs: dict[str, Any],
    cursor: str | None,
) -> dict[str, Any]:
    params = base_kwargs.copy()
    if cursor:
        params["start_cursor"] = cursor
    return params


def _apply_result_limit(
    results: list[Any], total_limit: int | None, total_fetched: int
) -> list[Any]:
    if total_limit is None:
        return results

    remaining_space = total_limit - total_fetched
    return results[:remaining_space]


def _has_reached_limit(total_limit: int | None, total_fetched: int) -> bool:
    if total_limit is None:
        return False
    return total_fetched >= total_limit


def _create_limited_response(
    original: PaginatedResponse,
    limited_results: list[Any],
    api_page_size: int,
) -> PaginatedResponse:
    results_were_limited_by_client = len(limited_results) < len(original.results)
    api_returned_full_page = len(original.results) == api_page_size

    has_more_after_limit = (
        original.has_more
        and not results_were_limited_by_client
        and api_returned_full_page
    )

    return PaginatedResponse(
        results=limited_results,
        has_more=has_more_after_limit,
        next_cursor=original.next_cursor if has_more_after_limit else None,
    )


async def paginate_notion_api(
    api_call: Callable[..., Coroutine[Any, Any, PaginatedResponse]],
    total_results_limit: int | None = None,
    **kwargs,
) -> list[Any]:
    all_results = []
    async for page in _fetch_data(
        api_call, total_results_limit=total_results_limit, **kwargs
    ):
        all_results.extend(page.results)
    return all_results


async def paginate_notion_api_generator(
    api_call: Callable[..., Coroutine[Any, Any, PaginatedResponse]],
    total_results_limit: int | None = None,
    **kwargs,
) -> AsyncGenerator[Any]:
    async for page in _fetch_data(api_call, total_results_limit, **kwargs):
        for item in page.results:
            yield item



from .query import NotionWorkspaceQueryConfigBuilder, WorkspaceQueryConfig
from .service import NotionWorkspace

__all__ = [
    "NotionWorkspace",
    "NotionWorkspaceQueryConfigBuilder",
    "WorkspaceQueryConfig",
]



from typing import override

from notionary.blocks.schemas import CreateEmbedBlock, EmbedData
from notionary.page.content.parser.parsers.base import BlockParsingContext, LineParser
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class EmbedParser(LineParser):
    def __init__(self, syntax_registry: SyntaxDefinitionRegistry) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_embed_syntax()

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._syntax.regex_pattern.search(context.line) is not None

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        url = self._extract_url(context.line)
        if not url:
            return

        embed_data = EmbedData(url=url, caption=[])
        block = CreateEmbedBlock(embed=embed_data)
        context.result_blocks.append(block)

    def _extract_url(self, line: str) -> str | None:
        match = self._syntax.regex_pattern.search(line)
        return match.group(1) if match else None



from __future__ import annotations

from collections.abc import AsyncIterator, Callable
from typing import TYPE_CHECKING, Self

from notionary.user.service import UserService
from notionary.workspace.query.builder import NotionWorkspaceQueryConfigBuilder
from notionary.workspace.query.models import (
    WorkspaceQueryConfig,
    WorkspaceQueryObjectType,
)
from notionary.workspace.query.service import WorkspaceQueryService

if TYPE_CHECKING:
    from notionary import NotionDataSource, NotionPage
    from notionary.user import BotUser, PersonUser


class NotionWorkspace:
    def __init__(
        self,
        name: str | None = None,
        query_service: WorkspaceQueryService | None = None,
        user_service: UserService | None = None,
    ) -> None:
        self._name = name
        self._query_service = query_service or WorkspaceQueryService()
        self._user_service = user_service or UserService()

    @classmethod
    async def from_current_integration(cls) -> Self:
        from notionary.user import BotUser

        bot_user = await BotUser.from_current_integration()

        return cls(name=bot_user.workspace_name)

    @property
    def name(self) -> str:
        return self._name

    async def get_pages(
        self,
        *,
        filter_fn: Callable[
            [NotionWorkspaceQueryConfigBuilder], NotionWorkspaceQueryConfigBuilder
        ]
        | None = None,
        query_config: WorkspaceQueryConfig | None = None,
    ) -> list[NotionPage]:
        if filter_fn is not None and query_config is not None:
            raise ValueError("Use either filter_fn OR query_config, not both")

        resolved_config = self._resolve_query_config(
            filter_fn, query_config, WorkspaceQueryObjectType.PAGE
        )
        return await self._query_service.get_pages(resolved_config)

    async def get_pages_stream(
        self,
        *,
        filter_fn: Callable[
            [NotionWorkspaceQueryConfigBuilder], NotionWorkspaceQueryConfigBuilder
        ]
        | None = None,
        query_config: WorkspaceQueryConfig | None = None,
    ) -> AsyncIterator[NotionPage]:
        if filter_fn is not None and query_config is not None:
            raise ValueError("Use either filter_fn OR query_config, not both")

        resolved_config = self._resolve_query_config(
            filter_fn, query_config, WorkspaceQueryObjectType.PAGE
        )
        async for page in self._query_service.get_pages_stream(resolved_config):
            yield page

    async def get_data_sources(
        self,
        *,
        filter_fn: Callable[
            [NotionWorkspaceQueryConfigBuilder], NotionWorkspaceQueryConfigBuilder
        ]
        | None = None,
        query_config: WorkspaceQueryConfig | None = None,
    ) -> list[NotionDataSource]:
        if filter_fn is not None and query_config is not None:
            raise ValueError("Use either filter_fn OR query_config, not both")

        resolved_config = self._resolve_query_config(
            filter_fn, query_config, WorkspaceQueryObjectType.DATA_SOURCE
        )
        return await self._query_service.get_data_sources(resolved_config)

    async def get_data_sources_stream(
        self,
        *,
        filter_fn: Callable[
            [NotionWorkspaceQueryConfigBuilder], NotionWorkspaceQueryConfigBuilder
        ]
        | None = None,
        query_config: WorkspaceQueryConfig | None = None,
    ) -> AsyncIterator[NotionDataSource]:
        if filter_fn is not None and query_config is not None:
            raise ValueError("Use either filter_fn OR query_config, not both")

        resolved_config = self._resolve_query_config(
            filter_fn, query_config, WorkspaceQueryObjectType.DATA_SOURCE
        )
        async for data_source in self._query_service.get_data_sources_stream(
            resolved_config
        ):
            yield data_source

    def _resolve_query_config(
        self,
        filter_fn: Callable[
            [NotionWorkspaceQueryConfigBuilder], NotionWorkspaceQueryConfigBuilder
        ]
        | None,
        query_config: WorkspaceQueryConfig | None,
        expected_object_type: WorkspaceQueryObjectType,
    ) -> WorkspaceQueryConfig:
        if filter_fn is not None:
            builder = NotionWorkspaceQueryConfigBuilder()
            configured_builder = filter_fn(builder)
            query_config = configured_builder.build()

        if query_config is None:
            query_config = WorkspaceQueryConfig(object_type=expected_object_type)
        else:
            query_config.object_type = expected_object_type

        return query_config

    async def get_users(self) -> list[PersonUser]:
        return [user async for user in self._user_service.list_users_stream()]

    async def get_users_stream(self) -> AsyncIterator[PersonUser]:
        async for user in self._user_service.list_users_stream():
            yield user

    async def get_bot_users(self) -> list[BotUser]:
        return [user async for user in self._user_service.list_bot_users_stream()]

    async def get_bot_users_stream(self) -> AsyncIterator[BotUser]:
        async for user in self._user_service.list_bot_users_stream():
            yield user

    async def search_users(self, query: str) -> list[PersonUser]:
        return [user async for user in self._user_service.search_users_stream(query)]

    async def search_users_stream(self, query: str) -> AsyncIterator[PersonUser]:
        async for user in self._user_service.search_users_stream(query):
            yield user
